from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Final,
    Generic,
    Optional,
    Union,
)

from koda._generics import A, B, FailT

if TYPE_CHECKING:  # pragma: no cover
    from koda.result import Result


class Nothing:
    # singleton
    _instance: ClassVar[Optional["Nothing"]] = None

    def __new__(cls) -> "Nothing":
        """
        Make `Nothing` a singleton, so we can do `is` checks if we want.
        """
        if cls._instance is None:
            cls._instance = super(Nothing, cls).__new__(cls)
        return cls._instance

    def __eq__(self, other: Any) -> bool:
        return other is self._instance

    def __repr__(self) -> str:
        return "Nothing()"

    def get_or_else(self, fallback: A) -> A:
        return fallback

    def map(self, _: Callable[[Any], B]) -> "Maybe[B]":
        return self

    def flat_map(self, _: Callable[[Any], "Maybe[B]"]) -> "Maybe[B]":
        return self

    def apply(self, _: "Maybe[Callable[[Any], B]]") -> "Maybe[B]":
        return self

    @property
    def to_optional(self) -> Optional[Any]:
        """
        Note that `Just[None]` will return None!
        """
        return None

    def to_result(self, fail_obj: FailT) -> "Result[Any, FailT]":
        from koda.result import Err

        return Err(fail_obj)


# just a pre-init-ed instance of nothing.
nothing: Final[Nothing] = Nothing()


class Just(Generic[A]):
    __match_args__ = ("val",)
    __slots__ = ("val",)

    def __init__(self, val: A) -> None:
        self.val: A = val

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Just) and other.val == self.val

    def __repr__(self) -> str:
        return f"Just({self.val})"

    def get_or_else(self, _: Any) -> A:
        return self.val

    def map(self, fn: Callable[[A], B]) -> "Maybe[B]":
        return Just(fn(self.val))

    def flat_map(self, fn: Callable[[A], "Maybe[B]"]) -> "Maybe[B]":
        return fn(self.val)

    def apply(self, container: "Maybe[Callable[[A], B]]") -> "Maybe[B]":
        if isinstance(container, Nothing):
            return nothing
        else:
            return Just(container.val(self.val))

    @property
    def to_optional(self) -> Optional[A]:
        return self.val

    def to_result(self, fail_obj: FailT) -> "Result[A, FailT]":
        from koda.result import Ok

        return Ok(self.val)


Maybe = Union[Just[A], Nothing]



from typing import Any, Callable, Generic, Union

from koda._generics import A, B, C, D, E


class First(Generic[A]):
    __slots__ = ("val",)
    __match_args__ = ("val",)

    def __init__(self, val: A) -> None:
        self.val: A = val

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, First) and other.val == self.val

    def __repr__(self) -> str:
        return f"First({self.val})"

    def map_first(self, fn: Callable[[A], B]) -> "First[B]":
        return First(fn(self.val))

    def map_second(self, fn: Callable[[Any], Any]) -> "First[A]":
        return self

    def map_third(self, fn: Callable[[Any], Any]) -> "First[A]":
        return self

    def map_fourth(self, fn: Callable[[Any], Any]) -> "First[A]":
        return self

    def map_fifth(self, fn: Callable[[Any], Any]) -> "First[A]":
        return self


class Second(Generic[A]):
    __slots__ = ("val",)
    __match_args__ = ("val",)

    def __init__(self, val: A) -> None:
        self.val: A = val

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Second) and other.val == self.val

    def __repr__(self) -> str:
        return f"Second({self.val})"

    def map_first(self, fn: Callable[[Any], Any]) -> "Second[A]":
        return self

    def map_second(self, fn: Callable[[A], B]) -> "Second[B]":
        return Second(fn(self.val))

    def map_third(self, fn: Callable[[Any], Any]) -> "Second[A]":
        return self

    def map_fourth(self, fn: Callable[[Any], Any]) -> "Second[A]":
        return self

    def map_fifth(self, fn: Callable[[Any], Any]) -> "Second[A]":
        return self


class Third(Generic[A]):
    __slots__ = ("val",)
    __match_args__ = ("val",)

    def __init__(self, val: A) -> None:
        self.val: A = val

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Third) and other.val == self.val

    def __repr__(self) -> str:
        return f"Third({self.val})"

    def map_first(self, fn: Callable[[Any], Any]) -> "Third[A]":
        return self

    def map_second(self, fn: Callable[[Any], Any]) -> "Third[A]":
        return self

    def map_third(self, fn: Callable[[A], B]) -> "Third[B]":
        return Third(fn(self.val))

    def map_fourth(self, fn: Callable[[Any], Any]) -> "Third[A]":
        return self

    def map_fifth(self, fn: Callable[[Any], Any]) -> "Third[A]":
        return self


class Fourth(Generic[A]):
    __slots__ = ("val",)
    __match_args__ = ("val",)

    def __init__(self, val: A) -> None:
        self.val: A = val

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Fourth) and other.val == self.val

    def __repr__(self) -> str:
        return f"Fourth({self.val})"

    def map_first(self, fn: Callable[[Any], Any]) -> "Fourth[A]":
        return self

    def map_second(self, fn: Callable[[Any], Any]) -> "Fourth[A]":
        return self

    def map_third(self, fn: Callable[[Any], Any]) -> "Fourth[A]":
        return self

    def map_fourth(self, fn: Callable[[A], B]) -> "Fourth[B]":
        return Fourth(fn(self.val))

    def map_fifth(self, fn: Callable[[Any], Any]) -> "Fourth[A]":
        return self


class Fifth(Generic[A]):
    __slots__ = ("val",)
    __match_args__ = ("val",)

    def __init__(self, val: A) -> None:
        self.val: A = val

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Fifth) and other.val == self.val

    def __repr__(self) -> str:
        return f"Fifth({self.val})"

    def map_first(self, fn: Callable[[Any], Any]) -> "Fifth[A]":
        return self

    def map_second(self, fn: Callable[[Any], Any]) -> "Fifth[A]":
        return self

    def map_third(self, fn: Callable[[Any], Any]) -> "Fifth[A]":
        return self

    def map_fourth(self, fn: Callable[[Any], Any]) -> "Fifth[A]":
        return self

    def map_fifth(self, fn: Callable[[A], B]) -> "Fifth[B]":
        return Fifth(fn(self.val))


Either = Union[First[A], Second[B]]
Either3 = Union[First[A], Second[B], Third[C]]
Either4 = Union[First[A], Second[B], Third[C], Fourth[D]]
Either5 = Union[First[A], Second[B], Third[C], Fourth[D], Fifth[E]]



from koda.either import (
    Either,
    Either3,
    Either4,
    Either5,
    Fifth,
    First,
    Fourth,
    Second,
    Third,
)
from koda.maybe import Just, Maybe, Nothing, nothing
from koda.result import Err, Ok, Result
from koda.utils import (
    Thunk,
    always,
    compose,
    identity,
    load_once,
    mapping_get,
    safe_try,
    thunkify,
    to_maybe,
    to_result,
)

__all__ = (
    "always",
    "Either",
    "Either3",
    "Either4",
    "Either5",
    "First",
    "Second",
    "Third",
    "Fourth",
    "Fifth",
    "Maybe",
    "Just",
    "Nothing",
    "nothing",
    "Result",
    "Ok",
    "Err",
    "compose",
    "identity",
    "mapping_get",
    "load_once",
    "safe_try",
    "to_maybe",
    "to_result",
    "Thunk",
    "thunkify",
)



from typing import TYPE_CHECKING, Any, Callable, Generic, Optional, Union

from koda._generics import A, B, FailT

if TYPE_CHECKING:  # pragma: no cover
    from koda.maybe import Maybe


class Ok(Generic[A]):
    __match_args__ = ("val",)
    __slots__ = ("val",)

    def __init__(self, val: A) -> None:
        self.val: A = val

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Ok) and other.val == self.val

    def __repr__(self) -> str:
        return f"Ok({self.val})"

    def apply(self, container: "Result[Callable[[A], B], FailT]") -> "Result[B, FailT]":
        if isinstance(container, Ok):
            return Ok(container.val(self.val))
        else:
            return container

    def get_or_else(self, _: A) -> A:
        return self.val

    def flat_map(self, fn: Callable[[A], "Result[B, FailT]"]) -> "Result[B, FailT]":
        return fn(self.val)

    def flat_map_err(self, fn: Callable[[Any], "Result[A, B]"]) -> "Result[A, B]":
        return self

    def map(self, fn: Callable[[A], B]) -> "Result[B, FailT]":
        return Ok(fn(self.val))

    def map_err(self, fn: Callable[[Any], "B"]) -> "Result[A, B]":
        return self

    def swap(self) -> "Result[FailT, A]":
        return Err(self.val)

    @property
    def to_optional(self) -> Optional[A]:
        """
        Note that `Ok[None]` will return None!
        """
        return self.val

    @property
    def to_maybe(self) -> "Maybe[A]":
        from koda.maybe import Just

        return Just(self.val)


class Err(Generic[FailT]):
    __match_args__ = ("val",)
    __slots__ = ("val",)

    def __init__(self, val: FailT) -> None:
        self.val: FailT = val

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Err) and other.val == self.val

    def __repr__(self) -> str:
        return f"Err({self.val})"

    def apply(self, _: "Result[Callable[[Any], B], FailT]") -> "Result[B, FailT]":
        return self

    def get_or_else(self, fallback: A) -> A:
        return fallback

    def map(self, _: Callable[[Any], B]) -> "Result[B, FailT]":
        return self

    def flat_map(self, _: Callable[[Any], "Result[B, FailT]"]) -> "Result[B, FailT]":
        return self

    def flat_map_err(self, fn: Callable[[FailT], "Result[A, B]"]) -> "Result[A, B]":
        return fn(self.val)

    def map_err(self, fn: Callable[[FailT], B]) -> "Result[A, B]":
        return Err(fn(self.val))

    def swap(self) -> "Result[FailT, A]":
        return Ok(self.val)

    @property
    def to_optional(self) -> Optional[Any]:
        return None

    @property
    def to_maybe(self) -> "Maybe[Any]":
        from koda.maybe import nothing

        return nothing


Result = Union[Ok[A], Err[FailT]]



from typing import Callable, Final, Optional, Union, cast, overload

from koda._generics import A, B, C, D, E, F, G, H, I
from koda.result import Err, Ok, Result


@overload
def _compose(fn1: Callable[[A], B], fn2: Callable[[B], C]) -> Callable[[A], C]:
    ...  # pragma: no cover


@overload
def _compose(
    fn1: Callable[[A], B], fn2: Callable[[B], C], fn3: Callable[[C], D]
) -> Callable[[A], D]:
    ...  # pragma: no cover


@overload
def _compose(
    fn1: Callable[[A], B],
    fn2: Callable[[B], C],
    fn3: Callable[[C], D],
    fn4: Callable[[D], E],
) -> Callable[[A], E]:
    ...  # pragma: no cover


@overload
def _compose(
    fn1: Callable[[A], B],
    fn2: Callable[[B], C],
    fn3: Callable[[C], D],
    fn4: Callable[[D], E],
    fn5: Callable[[E], F],
) -> Callable[[A], F]:
    ...  # pragma: no cover


@overload
def _compose(
    fn1: Callable[[A], B],
    fn2: Callable[[B], C],
    fn3: Callable[[C], D],
    fn4: Callable[[D], E],
    fn5: Callable[[E], F],
    fn6: Callable[[F], G],
) -> Callable[[A], G]:
    ...  # pragma: no cover


@overload
def _compose(
    fn1: Callable[[A], B],
    fn2: Callable[[B], C],
    fn3: Callable[[C], D],
    fn4: Callable[[D], E],
    fn5: Callable[[E], F],
    fn6: Callable[[F], G],
    fn7: Callable[[G], H],
) -> Callable[[A], H]:
    ...  # pragma: no cover


@overload
def _compose(
    fn1: Callable[[A], B],
    fn2: Callable[[B], C],
    fn3: Callable[[C], D],
    fn4: Callable[[D], E],
    fn5: Callable[[E], F],
    fn6: Callable[[F], G],
    fn7: Callable[[G], H],
    fn8: Callable[[H], I],
) -> Callable[[A], I]:
    ...  # pragma: no cover


def _compose(
    fn1: Callable[[A], B],
    fn2: Callable[[B], C],
    fn3: Optional[Callable[[C], D]] = None,
    fn4: Optional[Callable[[D], E]] = None,
    fn5: Optional[Callable[[E], F]] = None,
    fn6: Optional[Callable[[F], G]] = None,
    fn7: Optional[Callable[[G], H]] = None,
    fn8: Optional[Callable[[H], I]] = None,
) -> Callable[[A], Union[C, D, E, F, G, H, I]]:
    def inner(obj: A) -> Union[C, D, E, F, G, H, I]:
        if fn3 is None:
            return fn2(fn1(obj))
        elif fn4 is None:
            return fn3(fn2(fn1(obj)))
        elif fn5 is None:
            return fn4(fn3(fn2(fn1(obj))))
        elif fn6 is None:
            return fn5(fn4(fn3(fn2(fn1(obj)))))
        elif fn7 is None:
            return fn6(fn5(fn4(fn3(fn2(fn1(obj))))))
        elif fn8 is None:
            return fn7(fn6(fn5(fn4(fn3(fn2(fn1(obj)))))))
        else:
            return fn8(fn7(fn6(fn5(fn4(fn3(fn2(fn1(obj))))))))

    return inner


class _Unset:
    pass


_unset: Final[_Unset] = _Unset()


@overload
def _safe_try(fn: Callable[[A], B], v1: A) -> Result[B, Exception]:
    ...  # pragma: no cover


@overload
def _safe_try(fn: Callable[[A, B], C], v1: A, v2: B) -> Result[C, Exception]:
    ...  # pragma: no cover


@overload
def _safe_try(fn: Callable[[A, B, C], D], v1: A, v2: B, v3: C) -> Result[D, Exception]:
    ...  # pragma: no cover


@overload
def _safe_try(
    fn: Callable[[A, B, C, D], E], v1: A, v2: B, v3: C, v4: D
) -> Result[E, Exception]:
    ...  # pragma: no cover


@overload
def _safe_try(
    fn: Callable[[A, B, C, D, E], F], v1: A, v2: B, v3: C, v4: D, v5: E
) -> Result[F, Exception]:
    ...  # pragma: no cover


@overload
def _safe_try(
    fn: Callable[[A, B, C, D, E, F], G], v1: A, v2: B, v3: C, v4: D, v5: E, v6: F
) -> Result[G, Exception]:
    ...  # pragma: no cover


def _safe_try(
    fn: Union[
        Callable[[A], B],
        Callable[[A, B], C],
        Callable[[A, B, C], D],
        Callable[[A, B, C, D], E],
        Callable[[A, B, C, D, E], F],
        Callable[[A, B, C, D, E, F], G],
    ],
    v1: A,
    v2: Union[B, _Unset] = _unset,
    v3: Union[C, _Unset] = _unset,
    v4: Union[D, _Unset] = _unset,
    v5: Union[E, _Unset] = _unset,
    v6: Union[F, _Unset] = _unset,
) -> Union[
    Result[B, Exception],
    Result[C, Exception],
    Result[D, Exception],
    Result[E, Exception],
    Result[F, Exception],
    Result[G, Exception],
]:
    if isinstance(v2, _Unset):
        try:
            return Ok(cast(Callable[[A], B], fn)(v1))
        except Exception as e:
            return Err(e)
    elif isinstance(v3, _Unset):
        try:
            return Ok(cast(Callable[[A, B], C], fn)(v1, v2))
        except Exception as e:
            return Err(e)
    elif isinstance(v4, _Unset):
        try:
            return Ok(cast(Callable[[A, B, C], D], fn)(v1, v2, v3))
        except Exception as e:
            return Err(e)
    elif isinstance(v5, _Unset):
        try:
            return Ok(cast(Callable[[A, B, C, D], E], fn)(v1, v2, v3, v4))
        except Exception as e:
            return Err(e)
    elif isinstance(v6, _Unset):
        try:
            return Ok(cast(Callable[[A, B, C, D, E], F], fn)(v1, v2, v3, v4, v5))
        except Exception as e:
            return Err(e)
    else:
        try:
            return Ok(cast(Callable[[A, B, C, D, E, F], G], fn)(v1, v2, v3, v4, v5, v6))
        except Exception as e:
            return Err(e)



from typing import TypeVar

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
D = TypeVar("D")
E = TypeVar("E")
F = TypeVar("F")
G = TypeVar("G")
H = TypeVar("H")
I = TypeVar("I")
J = TypeVar("J")
K = TypeVar("K")

FailT = TypeVar("FailT")



from typing import Any, Callable, List, Mapping, Optional, Protocol, TypeVar

from koda._cruft import _compose, _safe_try
from koda._generics import A, B, FailT
from koda.maybe import Just, Maybe, nothing
from koda.result import Err, Ok, Result

compose = _compose


def identity(x: A) -> A:
    return x


def mapping_get(data: Mapping[A, B], key: A) -> Maybe[B]:
    # this is better than data.get(...) because if None is a valid return value,
    # there's no way to know if the value is the value from the map or the deafult value
    try:
        return Just(data[key])
    except KeyError:
        return nothing


def to_maybe(val: Optional[A]) -> Maybe[A]:
    if val is None:
        return nothing
    else:
        return Just(val)


def to_result(val: Optional[A], if_none: FailT) -> Result[A, FailT]:
    if val is None:
        return Err(if_none)
    else:
        return Ok(val)


Thunk = Callable[[], A]


def thunkify(obj: A) -> Thunk[A]:
    def inner() -> A:
        return obj

    return inner


def load_once(fn: Thunk[A]) -> Thunk[A]:
    """
    Lazily get some value
    """
    container: List[A] = []

    def inner() -> A:
        if len(container) == 0:
            val: A = fn()
            container.append(val)
            return val
        else:
            return container[0]

    return inner


safe_try = _safe_try


A_co = TypeVar("A_co", covariant=True)


class _AnyArgs(Protocol[A_co]):  # pragma: no cover
    """
    todo: consider removing, and just using Callable[..., A] below in `always`
    """

    def __call__(self, *args: Any, **kwargs: Any) -> A_co:
        ...


def always(x: A) -> _AnyArgs[A]:
    def inner(*args: Any, **kwargs: Any) -> A:
        return x

    return inner



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['koda']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'koda',
    'version': '1.3.1',
    'description': 'Type-safe functional tools for Python',
    'long_description': '# Koda\n\nKoda is a collection of practical type-safe tools for Python.\n\nAt its core are a number of datatypes that are common in functional programming.\n\n## Maybe\n\n`Maybe` is similar to Python\'s `Optional` type. It has two variants: `Nothing` and `Just`, and they work in similar ways\nto what you may have seen in other languages.\n\n```python3\nfrom koda import Maybe, Just, nothing\n\na: Maybe[int] = Just(5)\nb: Maybe[int] = nothing\n```\n\nTo know if a `Maybe` is a `Just` or a `Nothing`, you\'ll need to inspect it.\n\n```python3\nfrom koda import Just, Maybe\n\nmaybe_str: Maybe[str] = function_returning_maybe_str()\n\n# python 3.10 +\nmatch maybe_str:\n    case Just(val):\n        print(val)\n    case Nothing:\n        print("No value!")\n\n# python 3.9 and earlier\nif isinstance(maybe_str, Just):\n    print(maybe_str.val)\nelse:\n    print("No value!")\n```\n\n`Maybe` has methods for conveniently stringing logic together.\n\n#### Maybe.map\n\n```python3\nfrom koda import Just, nothing\n\ndef add_10(x: int) -> int:\n    return x + 10\n\n\nJust(5).map(add_10)  # Just(15)\nnothing.map(add_10)  # nothing \nJust(5).map(add_10).map(lambda x: f"abc{x}")  # Just("abc15")\n```\n\n#### Maybe.flat_map\n\n```python3\nfrom koda import Maybe, Just, nothing\n\n\ndef safe_divide(dividend: int, divisor: int) -> Maybe[float]:\n    if divisor != 0:\n        return Just(dividend / divisor)\n    else:\n        return nothing\n\nJust(5).flat_map(lambda x: safe_divide(10, x))  # Just(2)\nJust(0).flat_map(lambda x: safe_divide(10, x))  # nothing\nnothing.flat_map(lambda x: safe_divide(10, x))  # nothing\n```\n\n## Result\n\n`Result` provides a means of representing whether a computation succeeded or failed. To represent success, we can use `OK`;\nfor failures we can use `Err`. Compared to `Maybe`, `Result` is perhaps most useful in that the "failure" case also returns data,\nwhereas `Nothing` contains no data.\n\n```python3\nfrom koda import Ok, Err, Result \n\n\ndef safe_divide_result(dividend: int, divisor: int) -> Result[float, str]:\n    if divisor != 0:\n        return Ok(dividend / divisor)\n    else:\n        return Err("cannot divide by zero!")\n\n\nOk(5).flat_map(lambda x: safe_divide_result(10, x))  # Ok(2)\nOk(0).flat_map(lambda x: safe_divide_result(10, x))  # Err("cannot divide by zero!") \nErr("some other error").map(lambda x: safe_divide_result(10, x))  # Err("some other error")\n```\n\n`Result` can be convenient with `try`/`except` logic.\n```python3\nfrom koda import Result, Ok, Err\n\ndef divide_by(dividend: int, divisor: int) -> Result[float, ZeroDivisionError]:\n    try:\n        return Ok(dividend / divisor)\n    except ZeroDivisionError as exc:\n        return Err(exc)\n\n\ndivided: Result[float, ZeroDivisionError] = divide_by(10, 0)  # Err(ZeroDivisionError("division by zero"))\n```\n\nAnother way to perform the same computation would be to use `safe_try`:\n```python3\nfrom koda import Result, safe_try\n\n\n# not safe on its own!\ndef divide(dividend: int, divisor: int) -> float:\n    return dividend / divisor\n\n# safe if used with `safe_try`\ndivided_ok: Result[float, Exception] = safe_try(divide, 10, 2)  # Ok(5)\ndivided_err: Result[float, Exception] = safe_try(divide, 10, 0)  # Err(ZeroDivisionError("division by zero"))\n```\n\n### Conversion between `Result`s, `Maybe`s, and `Optional`s\n\n### Result and Maybe\n\nConvert a `Result` to a `Maybe` type.\n\n```python3\nfrom koda import Just, nothing, Ok, Err\n\nassert Ok(5).to_maybe == Just(5)\nassert Err("any error").to_maybe == nothing \n```\n\nConvert a `Maybe` to a `Result` type.\n\n```python3\nfrom koda import Just, nothing, Ok, Err\n\nassert nothing.to_result("value if nothing") == Err("value if nothing")\nassert Just(5).to_result("value if nothing") == Ok(5)\n```\n\n### `Maybe` and `Optional`\n\nConvert an `Optional` value to a `Maybe`.\n\n```python3\nfrom koda import to_maybe, Just, nothing\n\nassert to_maybe(5) == Just(5)\nassert to_maybe("abc") == Just("abc")\nassert to_maybe(False) == Just(False)\n\nassert to_maybe(None) == nothing\n```\n\nConvert a `Maybe` to an `Optional`.\n```python3\nfrom koda import Just, nothing\n\nassert Just(5).to_optional == 5\nassert nothing.to_optional is None\n\n# note that `Maybe[None]` will always return None, \n# so `Maybe.get_or_else` would be preferable in this case\nassert Just(None) is None\n```\n\n### `Result` and `Optional`\n\nConvert an `Optional` value to a `Result`.\n\n```python3\nfrom koda import to_result, Ok, Err \n\nassert to_result(5, "fallback") == Ok(5)\nassert to_result("abc", "fallback") == Ok("abc")\nassert to_result(False, "fallback") == Ok(False)\n\nassert to_result(None, "fallback") == Err("fallback")\n\n```\n\nConvert a `Result` to an `Optional`.\n```python3\nfrom koda import Ok, Err\n\nassert Ok(5).to_optional == 5\nassert Err("some error").to_optional is None\n\n# note that `Result[None, Any]` will always return None, \n# so `Result.get_or_else` would be preferable in this case\nassert Ok(None) is None\n```\n\n## More\n\nThere are many other functions and datatypes included. Some examples:\n\n### compose\nCombine functions by sequencing.\n\n```python3\nfrom koda import compose\nfrom typing import Callable\n\ndef int_to_str(val: int) -> str:\n    return str(val)\n\ndef prepend_str_abc(val: str) -> str:\n    return f"abc{val}"    \n\ncombined_func: Callable[[int], str] = compose(int_to_str, prepend_str_abc)\nassert combined_func(10) == "abc10"\n```\n\n### mapping_get\nTry to get a value from a `Mapping` object, and return an unambiguous result.\n\n```python3\nfrom koda import mapping_get, Just, Maybe, nothing\n\nexample_dict: dict[str, Maybe[int]] = {"a": Just(1), "b": nothing}\n\nassert mapping_get(example_dict, "a") == Just(Just(1))\nassert mapping_get(example_dict, "b") == Just(nothing)\nassert mapping_get(example_dict, "c") == nothing\n```\n\nAs a comparison, note that `dict.get` can return ambiguous results:\n```python\nfrom typing import Optional\n\nexample_dict: dict[str, Optional[int]] = {"a": 1, "b": None}\n\nassert example_dict.get("b") is None\nassert example_dict.get("c") is None\n```\nWe can\'t tell from the resulting value whether the `None` was the \nvalue for a key, or whether the key was not present in the `dict`\n\n### load_once\nCreate a lazy function, which will only call the passed-in function\nthe first time it is called. After it is called, the value is cached.\nThe cached value is returned on each successive call.\n```python3\nfrom random import random\nfrom koda import load_once\n\ncall_random_once = load_once(random)  # has not called random yet\n\nretrieved_val: float = call_random_once()\nassert retrieved_val == call_random_once()\n```\n\n## Intent\n\nKoda is intended to focus on a small set of practical data types and utility functions for Python. It will not \ngrow to encompass every possible functional or typesafe concept. Similarly, the intent of this library is to avoid \nrequiring extra plugins (beyond a type-checker like mypy or pyright) or specific typchecker settings. As such,\nit is unlikely that things like Higher Kinded Types emulation or extended type inference will be implemented in this \nlibrary.\n',
    'author': 'Keith Philpott',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/keithasaurus/koda',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
